(ns com.aurellem.gb.moves
  (:use (com.aurellem.gb gb-driver util constants hxc))
  (:import [com.aurellem.gb.gb_driver SaveState]))


;;; grab the move data directly from the rom
(def move-code->move-name ;; alternate name: moves
  ((comp vec cons)
   :end-of-moves
   (map format-name (hxc-move-names)))))

(def move-name->move-code ;; alternate name: move-id
  (zipmap
   move-names
   (range)))

(def move-name->move-pp
  (comp :pp (hxc-move-data)))

(defn max-pp [name pp-ups]
  (if (= 40 (move-name->move-pp name))
    (+ 40 (* 7 pp-ups))
    (int (* (+ 1 (* (/ 1 5) pp-ups))
            (move-name->move-pp name)))))

(def moves-codes-pokemon-1 0xD172)

(defn moves-codes-start [pokemon-num]
  (assert (<= 0 pokemon-num 5))
  (+ moves-codes-pokemon-1
     (* pokemon-num pokemon-record-width)))

(defn read-moves
  ([^SaveState state poke-num]
     (let [start (moves-codes-start poke-num)]
       (vec
        (take-while
         (partial not= :end-of-moves)
         (map
          move-code->move-name
          (subvec (vec (memory state))
                  start (+ start 4)))))))
  ([poke-num]
     (read-moves @current-state poke-num)))
             
(defn give-moves
  ([^SaveState state pokemon-num moves]
     (assert (<= (count moves) 4))
     (set-memory-range
      state
      (moves-codes-start pokemon-num)
      (map #(move-name->move-code % %)
           (concat moves
                   (repeat (- 4 (count moves)) :end-of-moves)))))
  ([pokemon-num moves]
     (give-moves @current-state pokemon-num moves))
  ([moves]
     (give-moves 0 moves)))

;; Note regarding PP of moves -- both the current PP and the
;; total PP are stored in the same value.
;; they are bit-packed, with the first 2 bits containing the
;; number of pp-ups that have been applied, and the next
;; six bits containing the current pp of the move.
;; thus, a move can have up to 63 current pp and up to
;; three pp-ups applied.

(def pokemon-1-pp-start 0xD187)

(defn moves-pp-start [pokemon-num]
  (assert (<= 0 pokemon-num 5))
  (+ pokemon-1-pp-start (* pokemon-num pokemon-record-width)))

(defn read-pp
  ([^SaveState state pokemon-num move-num]
     (assert (<= 0 move-num 3))
     (assert (<= 0 pokemon-num 5))
     (let [pp-raw
           (aget (memory state)
                 (+ (moves-pp-start pokemon-num)
                    move-num))
           pp-up
           (bit-shift-right
            (bit-and
             pp-raw
             (Integer/parseInt "11000000" 2)) 6)
           current-pp
           (bit-and
            pp-raw
            (Integer/parseInt "00111111" 2))]
       {:pp-ups pp-up :current-pp current-pp}))
  ([pokemon-num move-num]
     (read-pp @current-state pokemon-num move-num)))

(defn give-pp
  ([^SaveState state pokemon-num move-num pp-ups current-pp]
     (assert (<= 0 move-num 3))
     (assert (<= 0 pokemon-num 5))
     (assert (<= 0 pp-ups 3))
     (assert (<= 0 current-pp 63))

     (set-memory
      state
      (+ (moves-pp-start pokemon-num)
         move-num)
      (+
       (bit-shift-left pp-ups 6)
       (bit-and (Integer/parseInt
                 "00111111" 2)
                current-pp))))
  ([pokemon-num move-num pp-ups current-pp]
     (give-pp @current-state
             pokemon-num move-num pp-ups current-pp)))

(defn give-moves-pps
  ([^SaveState state poke-num moves-pps]
     (let [new-moves
           (-> state
               ;; zero out the pp of the pokemon's moves
               (give-pp poke-num 0 0 0)
               (give-pp poke-num 1 0 0)
               (give-pp poke-num 2 0 0)
               (give-pp poke-num 3 0 0)
               (give-moves poke-num (map first moves-pps)))]
       (reduce (fn [state move-num]
                 (let [pp (second (nth moves-pps move-num))]
                   (give-pp state poke-num move-num
                            (:pp-ups pp)
                            (:current-pp    pp))))
               new-moves (range (count moves-pps)))))
  ([poke-num moves-pps]
     (give-moves-pps @current-state poke-num moves-pps)))
           

